#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-11-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : compress

"""
Compresses an array between values by normalizing, with possibility to saturate extrema.
"""



# %% Libraries
import numpy as np
from arrlp import get_xp



# %% Function
def compress(array, /, max=1, min=0, *, dtype=None, white=None, black=None, white_percent=None, black_percent=None, saturate=None) :
    '''
    Compresses an array between values by normalizing, with possibility to saturate extrema.
    
    Parameters
    ----------
    array : np.ndarray
        Array to normalize.
    max : int or float or None
        white value in output. None for no changing of white
    min : int or float or None
        black value in output. None for no changing of black
    dtype : np.dtype or str or None
        dtype of output, None for same as input
    white : int or float or None
        white value in input. None for maximum
    black : int or float or None
        black value in input. None for minimum
    white_percent : int or None
        white percentage distribution in input if white is None.
    black_percent : int or None
        black percentage distribution in input if black is None.
    saturate : Any or bool or None
        If True, will saturate values above white and black. If Any, will replace by this value. If None no saturation.

    Returns
    -------
    array : np.ndarray
        Normalized and saturated copy of array.

    Examples
    --------
    >>> from arrlp import compress
    >>> array = np.arange(100, dtype=np.float32)
    ...
    >>> compress(array, 10, 5) # compresses array between 10 and 5
    >>> compress(array, white=50, black=40) # compresses array linearly so that 50 value is at 1 and 40 is at 0
    >>> compress(array, white=50, black=40, saturate=np.nan) # compresses array linearly so that 50 value is at 1 and 40 is at 0, replace outside values by np.nan
    >>> compress(array, white_percent=10, black=1) # compresses array linearly so that 10% of array will be white, and 1% black (without saturation)
    >>> compress(array, white_percent=10, black=1, saturate=True) # compresses array linearly so that 10% of array will be white, and 1% black (with saturation)
    '''

    # init
    xp = get_xp(array)
    array = xp.asarray(array)
    if dtype is None :
        dtype = array.dtype

    # Get white/black
    if white is None :
        white = xp.nanmax(array) if white_percent is None else xp.nanpercentile(array, 100-white_percent)
    if black is None :
        black = xp.nanmin(array) if black_percent is None else xp.nanpercentile(array, black_percent)
    if white <= black :
        raise ValueError('white <= black is not possible while compressing')

    # Normalization
    if max is not None and min is not None and min >= max :
        raise ValueError('min >= max is not possible while compressing')
    if max is not None :
        array = normalization(array, value=max, norm=white, fix=black, xp=xp)
        white = max
    if min is not None :
        array = normalization(array, value=min, norm=black, fix=white, xp=xp)
        black = min
    if max is None and min is None :
        array = xp.copy(array)

    # Saturation
    if saturate is not None :
        if saturate is True :
            sat_min, sat_max = black, white
        else :
            sat_min, sat_max = saturate, saturate
        array[array>white] = sat_max
        array[array<black] = sat_min
    
    return array.astype(dtype)



def normalization(array, /, value:float=None, norm:float=None, fix:float=None, xp=np):
    '''Basic normalization process of array copy while keeping a fixed point'''

    if fix is None : fix = 0
    if norm is None : norm = max(xp.nanmax(array),-xp.nanmin(array))
    if value is None : value = 1*xp.sign(norm)
    return (array-fix)/(norm-fix)*(value-fix) + fix



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def ini_img_correlate1d(self, array, *, kernel, **kwargs) :
    return dict(
        kernel=(self.xp.asarray(kernel[0]), self.xp.asarray(kernel[1])), # y, x
    )

def out_img_correlate1d(self, array, **kwargs) :
    return self.xp.empty_like(array, dtype=self.xp.float32)

def cpu_img_correlate1d(self, out, array, *, kernel, mode='nearest', **kwargs) :
    return self.ndimage.correlate1d(self.ndimage.correlate1d(array, weights=kernel[0], output=out, axis=self.axes[0], mode=mode), weights=kernel[1], output=out, axis=self.axes[1], mode=mode, **kwargs)

def par_img_correlate1d(self, array, *, kernel, mode='nearest', **kwargs) :
    return self.ndimage.correlate1d(self.ndimage.correlate1d(array, weights=kernel[0], axis=0, mode=mode), weights=kernel[1], axis=1, mode=mode, **kwargs)

def gpu_img_correlate1d(self, out, array, *, kernel, mode='nearest', **kwargs) :
    return self.ndimage.correlate1d(self.ndimage.correlate1d(array, weights=kernel[0], axis=self.axes[0], mode=mode), weights=kernel[1], axis=self.axes[1], mode=mode, output=out, **kwargs) # First call must be allocated into a temporary buffer (output=None)



# Main function
img_correlate1d = FunctionArray(
    
    # Mandatory
    ndims = 2,
    cpu_function = cpu_img_correlate1d,
    par_function = par_img_correlate1d,
    gpu_function = gpu_img_correlate1d,
    out_function = out_img_correlate1d,
    ini_function = ini_img_correlate1d,

    # Loops
    par_loop = True,
    use_joblib = True, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    from arrlp import kernel
    func = img_correlate1d

    # Arguments
    kwargs = dict(
        kernel=(kernel(ndims=1, pixel=1, sigma=2), kernel(ndims=1, pixel=1, sigma=3)),
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-12
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : debug_array

"""
This function allows to debug array functions while comparing various modes.
"""



# %% Libraries
import numpy as np
try :
    import cupy as cp
except ImportError :
    cp = None
from time import perf_counter
import warnings



# %% Function
def debug_array(func, modes, stacks_list=None, channels_list=None, narrays=1, **kwargs) :
    f'''
    This function allows to debug array functions while comparing various modes.
    
    Parameters
    ----------
    func : function
        function to test.
    modes : dict
        dict of dicts corresponding to various modes that should give same results.
    stacks_list : list
        List of stacks bools to test, must same size as channels_list.
    channels_list : list
        List of channels bools to test, must same size as stacks_list.
    narrays : int
        Number of input arrays.
    **kwargs : dict
        constant keyword arguments.

    Examples
    --------
    >>> from arrlp import debug_array
    ...
    >>> debug_array(func, ndimes, dict("Normal": dict()))
    '''



    # Data size ~2**24
    nstacks = int(2**8)
    nchannels = int(2**2)
    match func.ndims :
        case 1:
            shape = (int(2**14),)
        case 2:
            shape = (int(2**7), int(2**7))
        case 3:
            shape = (int(2**4), int(2**5), int(2**5))
        case _: raise ValueError(f'number of dimensions cannot be {func.ndims}')

    # Manage modes
    modes = {key: value for key, value in modes.items() if cp is not None or not value["cuda"]}



    # Timeit function
    def timeit(key, arrays, stacks, channels, parallel, cuda, **kw) :

        # Init
        xp = cp if cuda else np
        arrays = [xp.asarray(array) for array in arrays]

        # Calculate
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            print('Compile run...', end='')
            func(*arrays, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, test=True, **kwargs, **kw)
            print('\rRun...        ', end='')
            tic = perf_counter()
            out = func(*arrays, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, test=True, **kwargs, **kw)
            toc = perf_counter()
            print(f'\r{key} took {(toc-tic)*1000:.3f}ms      ')
            if cuda : out = xp.asnumpy(out)
        return out



    # Loops
    if channels_list is None : channels_list = [False, False, True, True]
    if stacks_list is None : stacks_list = [False, True, False, True]
    for channels, stacks in zip(channels_list, stacks_list) :

        print(f'\nTesting for stacks={stacks} and channels={channels}')

        # Array
        match stacks, channels :
            case (True, True) :
                _shape = (nstacks, *shape, nchannels)
            case(True, False) :
                _shape = (nstacks, *shape)
            case(False, True) :
                _shape = (*shape, nchannels)
            case(False, False) :
                _shape = shape
        arrays = [np.random.random(_shape) for _ in range(narrays)]

        # Calculate
        results = {}
        ref = None
        for key, value in modes.items():
            parallel = value['parallel']
            if (not channels and not stacks) and parallel and hasattr(func, "use_joblib") and func.use_joblib : continue
            results[key] = timeit(key, arrays, stacks, channels, **value)
            if ref is None :
                ref = key
        
        # Compare
        for key, value in results.items():
            if key == ref : continue
            print(f"\rChecking correctness vs reference: {key}" + " "*(60-34-len(key)), end='')
            np.testing.assert_allclose(
                results[ref], results[key],
                rtol=1e-4,
                atol=1e-5,
                err_msg="Outputs differ between optimizations"
            )
        print('\r'+' '*60+'\n')



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import kernel, vol_correlate1d



# %% Function
def vol_gaussianfilter(array, *, sigma:float, pixel:float=1., out=None, # Arrays
        stacks=False, channels=False, parallel=False, cuda=False, test=False, iterator=range, # Modes
        **kwargs) :

    try :
        if len(pixel) != 3 :
            raise ValueError('Argument does not have good number of dimensions')
    except TypeError :
        pixel = (pixel, pixel, pixel)
    try :
        if len(sigma) != 3 :
            raise ValueError('Argument does not have good number of dimensions')
    except TypeError :
        sigma = (sigma, sigma, sigma)
    k = kernel(ndims=1, pixel=pixel[0], sigma=sigma[0]), kernel(ndims=1, pixel=pixel[1], sigma=sigma[1]), kernel(ndims=1, pixel=pixel[2], sigma=sigma[2])
    return vol_correlate1d(array, kernel=k, out=out, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, test=test, iterator=iterator, **kwargs)
vol_gaussianfilter.ndims = 3



if __name__ == '__main__' :
    from arrlp import debug_array
    func = vol_gaussianfilter

    # Arguments
    kwargs = dict(
        sigma=3.,
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-06
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : gc

"""
Makes garbage collection to free (V)RAM.
"""



# %% Libraries
import gc as garbage_collect
try :
    import cupy as cp
except ImportError :
    cp = None



# %% Function
def gc() :
    '''
    Makes garbage collection to free (V)RAM.
    
    Examples
    --------
    >>> from arrlp import gc
    ...
    >>> gc()
    '''

    garbage_collect.collect()
    if cp is not None :
        cp.get_default_memory_pool().free_all_blocks()
        cp.get_default_pinned_memory_pool().free_all_blocks()



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-04
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : nb_threads

"""
This context manager defines the number of logical cores on numba parallel threads.
"""



# %% Libraries
from contextlib import contextmanager
from numba import set_num_threads, get_num_threads
import os



# %% Function
@contextmanager
def nb_threads(parallel=False) :
    '''
    This context manager defines the number of logical cores on numba parallel threads.
    
    Parameters
    ----------
    threads : int or bool
        If bool set -1 by default for True and 1 for False
        If int sets the number of threads, -1 means maximum

    Examples
    --------
    >>> from arrlp import nb_threads
    ...
    >>> with nb_threads(3) : # 3 threads
    ...     parallel_funct()
    '''

    if parallel is False :
        threads = 1
    elif parallel is True :
        threads = os.cpu_count() or 1
    elif parallel == -1 :
        threads = os.cpu_count() or 1
    else :
        threads = parallel
    old = get_num_threads()
    set_num_threads(threads)
    try:
        yield
    finally:
        set_num_threads(old)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-17
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : relabel

"""
This file allows to test relabel

relabel : This function redefines a label array to fill potential holes inside the considered labels.
"""



# %% Libraries
from corelp import debug
from arrlp import relabel
from time import perf_counter
import numpy as np
try :
    import cupy as cp
except ImportError :
    cp = None
debug_folder = debug(__file__)



# %% Function test
def test_function() :
    '''
    Test relabel function
    '''
    n = 100000

    labels = np.arange(0, n*10 + 10, 10)
    relabel(labels)
    labels = np.arange(0, n*10 + 10, 10)/2
    tic = perf_counter()
    newlabels = relabel(labels)
    toc = perf_counter()
    print(f'CPU took {(toc-tic)*1000:.3f}ms')
    assert (newlabels == np.arange(n+1)).all()

    if cp is not None :
        labels = cp.arange(0, n*10 + 10, 10)
        relabel(labels)
        labels = cp.arange(0, n*10 + 10, 10)/2
        tic = perf_counter()
        newlabels = relabel(labels)
        toc = perf_counter()
        print(f'GPU took {(toc-tic)*1000:.3f}ms')
        assert (newlabels == cp.arange(n+1)).all()

    



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import arr_function, get_xp



# %% Function
def arr_wrapper(array, *, out=None, # Arrays
        stacks=False, channels=False, parallel=False, cuda=False, test=False, iterator=range, # Modes
        **kwargs) :

    xp = get_xp(cuda)
        
    return arr_function(array, out=out, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, test=test, iterator=iterator, **kwargs)
arr_wrapper.ndims = 2



if __name__ == '__main__' :
    from arrlp import debug_array
    func = arr_wrapper

    # Arguments
    kwargs = dict(
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-13
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : FunctionArray

"""
This class defines a function for various array configurations.
"""



# %% Libraries
import types
from dataclasses import dataclass, field
import numpy as np
import scipy
from joblib import Parallel, delayed
from numba import njit, prange
import skimage

# Check cupy
try :
    import cupy as cp
    import cupyx.scipy as cupyx_scipy
    from cupyx.scipy import ndimage as cupyx_ndimage
    import cucim.skimage as cucim_skimage
except ImportError :
    cp = None



# %% Class
@dataclass(slots=True, kw_only=True)
class FunctionArray() :
    '''
    This class defines a function for various array configurations.
    
    Parameters
    ----------
    ndims : int
        Number of dimensions of the base array.
    cpu_function : types.FunctionType
        Cpu function.
    par_function : types.FunctionType
        Par function.
    gpu_function : types.FunctionType
        Gpu function.
    out_function : types.FunctionType
        Function defining output array.
    ini_function : types.FunctionType
        Function defining initialization kwargs.
    cpu_loop : bool
        True if cpu_function applyies default loop.
    par_loop : bool
        True if par_function applyies default loop.
    gpu_loop : bool
        True if gpu_function applyies default loop.
    use_joblib : bool
        True to use joblib for parallel processes in parallel loop.
    remove_parallel : bool
        True if the parallel implementation is slower than normal python.
    remove_cuda : bool
        True if the cuda implementation is slower than normal python.

    Examples
    --------
    >>> from arrlp import FunctionArray
    ...
    >>> instance = FunctionArray(TODO)
    '''

    # Mandatory
    ndims : int
    cpu_function : types.FunctionType = field(repr=False)
    par_function : types.FunctionType = field(repr=False)
    gpu_function : types.FunctionType = field(repr=False)
    out_function : types.FunctionType = field(repr=False)
    ini_function : types.FunctionType = field(repr=False)

    # Loops
    cpu_loop : types.FunctionType = field(default=False, repr=False)
    par_loop : types.FunctionType = field(default=False, repr=False)
    gpu_loop : types.FunctionType = field(default=False, repr=False)
    use_joblib : bool = field(default=True, repr=False)

    # Performances
    remove_parallel : bool = field(default=False, repr=False)
    remove_cuda : bool = field(default=False, repr=False)



    # Methods
    stacks : bool = field(init=False, repr=False)
    channels : bool = field(init=False, repr=False)
    parallel : bool = field(init=False, repr=False)
    cuda : bool = field(init=False, repr=False)
    def __call__(self, array, *args, out=None, # Arrays
        stacks=False, channels=False, parallel=False, cuda=False, test=False, iterator=range, # Modes
        **kwargs) : # Arguments of the function
        '''
        This is the main function call
    
        Parameters
        ----------
        array : np.array
            At least one array from which we derive stacks and channels.
        *args : tuple(np.array)
            Tuple of array corresponding to inputs that follow the same stacks and channels dimensions.
        out : np.array
            Output array where to put result, might not be always available
        stacks : bool
            True to consider stacks dimensions
        channels : bool
            True to consider channels dimensions
        parallel : bool
            True to optimize in parallel cpu cores
        cuda : bool
            True to optimize on gpu
        test : bool
            [dev only], True when testing the speed to avoid raising errors on function that were defined as not optimal
        iterator : iterator
            Iterator defining how to loop on dimensions, must take an int as only input, allow to put progress bars.
        **kwargs : dict
            Other constant inputs to apply on each array

        Returns
        -------
        out : np.array
            Output array.
        '''

        # checks
        if iterator is None : iterator = range
        if cuda : parallel = False
        if parallel is True : parallel = -1
        if parallel == 1 : parallel = False
        self.checks(out, stacks, channels, parallel, cuda, test)
        self.stacks, self.channels, self.parallel, self.cuda = stacks, channels, parallel, cuda
        array = self.xp.asarray(array)
        args = [self.xp.asarray(arg) for arg in args]
        if self.ini_function is not None :
            kwargs.update(self.ini_function(self, array, *args, **kwargs))
        out = out if out is not None else self.out_function(self, array, *args, **kwargs) if self.out_function is not None else None

        # Cuda
        if cuda :
            if self.gpu_loop :
                return self.loop(self.gpu_function, iterator, out, array, *args, **kwargs)
            return self.gpu_function(self, out, array, *args, **kwargs)

        # Parallel
        elif parallel :
            if self.par_loop :
                if self.use_joblib :
                    return self.parallel_loop(self.par_function, out, array, *args, **kwargs)
                else :
                    return self.loop(self.par_function, iterator, out, array, *args, **kwargs)
            return self.par_function(self, out, array, *args, **kwargs)

        # Python
        else :
            if self.cpu_loop :
                return self.loop(self.cpu_function, iterator, out, array, *args, **kwargs)
            return self.cpu_function(self, out, array, *args, **kwargs)



    # Properties
    @property
    def axes(self) :
        start = int(self.stacks)
        return tuple(range(start, start + self.ndims))
    @property
    def xp(self) :
        return cp if self.cuda else np
    @property
    def scipyx(self) :
        return cupyx_scipy if self.cuda else scipy
    @property
    def ndimage(self) :
        return cupyx_ndimage if self.cuda else scipy.ndimage
    @property
    def skimagex(self) :
        return cucim_skimage if self.cuda else skimage
    def shape(self, array) : # function of array
        start = int(self.stacks)
        return array.shape[start:start + self.ndims]



    def checks(self, out, stacks, channels, parallel, cuda, test) :
        '''
        Make checks on asked mode
        '''

        # One optimization
        if parallel and cuda :
            raise ValueError('Cuda and Parallel cannot be True at the same time')
        
        # Cuda not available
        if cuda and cp is None :
            raise ValueError('Cuda was asked but is not available in this environment')

        # No parallel
        if parallel and self.remove_parallel and not test :
            raise ValueError('Parallel optimization is not effective in this function')

        # No cuda
        if cuda and self.remove_cuda and not test :
            raise ValueError('Cuda optimization is not effective in this function')

        # Joblib when no additional channels 
        if parallel and not stacks and not channels and self.use_joblib :
            raise ValueError('Normal array (no stack of channel) cannot be calculated in parallel')

        # Inplace out not possible
        if out is not None and self.out_function is None and not stacks and not channels :
            raise ValueError('Output cannot be defined in this function for normal single array')



    def loop(self, func, iterator, out, array, *args, **kwargs) :
    
        nstacks, nchannels = array.shape[0], array.shape[-1]
        match (self.stacks, self.channels) :

            case (False, False) :
                return func(self, out, array, *args, **kwargs)
            case (True, False) :
                for i in iterator(nstacks) :
                    stack_out = None if out is None else out[i]
                    _stack_out = func(self, stack_out, array[i], *(arg[i] for arg in args), **kwargs)
                    if out is None :
                        if i == 0 : _out = self.xp.empty_like(_stack_out, shape=(nstacks, *_stack_out.shape))
                        _out[i] = _stack_out
                return _out if out is None else out
            case (False, True) :
                for j in iterator(nchannels) :
                    channel_out = None if out is None else out[..., j]
                    _channel_out = func(self, channel_out, array[..., j], *(arg[..., j] for arg in args), **kwargs)
                    if out is None :
                        if j == 0 : _out = self.xp.empty_like(_channel_out, shape=(*_channel_out.shape, nchannels))
                        _out[..., j] = _channel_out
                return _out if out is None else out
            case (True, True) :
                for i in iterator(nstacks) :
                    stack_out, stack_array, stack_args = None if out is None else out[i], array[i], (arg[i] for arg in args)
                    if out is None and i > 0 : _stack_out = _out[i]
                    for j in range(nchannels) :
                        channel_out = None if stack_out is None else stack_out[..., j]
                        _channel_out = func(self, channel_out, stack_array[..., j], *(arg[..., j] for arg in stack_args), **kwargs)
                        if out is None :
                            if i == 0 and j == 0 : 
                                _out = self.xp.empty_like(_channel_out, shape=(nstacks, *_channel_out.shape, nchannels))
                                _stack_out = _out[i]
                            _stack_out[..., j] = _channel_out
                return _out if out is None else out

            case _ : raise SyntaxError(f'Cannot use (stacks, channels, out_name)={(self.stacks, self.channels, out_name)}')



    def parallel_loop(self, func, out, array, *args, **kwargs) :

        nstacks, nchannels = array.shape[0], array.shape[-1]
        match (self.stacks, self.channels) :

            case (False, False) :
                raise SyntaxError('This parallel scenario with no stack nor channel should not exist. [should be corrected in checks]')
            case (True, False) :
                copy = Parallel(n_jobs=self.parallel, backend="loky")(delayed(func)(self, array[i], *(arg[i] for arg in args), **kwargs) for i in range(nstacks))
                if out is None : out = np.empty_like(copy[0], shape=(len(copy), *copy[0].shape))
                return copystacks(list(copy), out)
            case (False, True) :
                copy = Parallel(n_jobs=self.parallel, backend="loky")(delayed(func)(self, array[..., i], *(arg[..., i] for arg in args), **kwargs) for i in range(nchannels))
                if out is None : out = np.empty_like(copy[0], shape=(*copy[0].shape, len(copy)))
                return copychannels(list(copy), out)
            case (True, True) :
                newfunc = lambda self, array, *args, **kwargs : [func(self, array[..., j], *(arg[..., j] for arg in args), **kwargs) for j in range(nchannels)]
                copy = Parallel(n_jobs=self.parallel, backend="loky")(delayed(newfunc)(self, array[i], *(arg[i] for arg in args), **kwargs) for i in range(nstacks))
                if out is None : out = np.empty_like(copy[0][0], shape=(len(copy), *copy[0][0].shape, len(copy[0])))
                return copystacksnchannels(list(copy), out)
            
            case _ : raise SyntaxError(f'Cannot use (stacks, channels)={(self.stacks, self.channels)}')



@njit(parallel=True)
def copystacks(copyfrom, copyto):
    for i in prange(len(copyfrom)):
        copyto[i] = copyfrom[i]
    return copyto

@njit(parallel=True)
def copychannels(copyfrom, copyto):
    for j in prange(len(copyfrom)):
        copyto[..., j] = copyfrom[j]
    return copyto

def copystacksnchannels(copyfrom, copyto):
    for i in prange(len(copyfrom)):
        stack_copyfrom, stack_copyto = copyfrom[i], copyto[i]
        for j in range(len(stack_copyfrom)) :
            stack_copyto[..., j] = stack_copyfrom[j]
    return copyto



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def _vol_crosscorr(self, out, array, array2) :
    array  = array.copy()
    array2 = array2.copy()
    
    array  -= array.mean(axis=self.axes, keepdims=True)
    array2 -= array2.mean(axis=self.axes, keepdims=True)
    array  /= array.std(axis=self.axes, keepdims=True)
    array2 /= array2.std(axis=self.axes, keepdims=True)
    
    array  = self.scipyx.fft.rfftn(array,  axes=self.axes) # release previous array
    array2 = self.scipyx.fft.rfftn(array2, axes=self.axes) # release previous array2
    self.xp.conj(array2, out=array2)
    array *= array2
    del(array2) # release previous array 2
    array = self.scipyx.fft.irfftn(array, axes=self.axes) # release previous array    
    return self.scipyx.fft.fftshift(array, axes=self.axes)

def par_vol_crosscorr(self, out, array, array2) :
    array  = array.copy()
    array2 = array2.copy()
    
    array  -= array.mean(axis=self.axes, keepdims=True)
    array2 -= array2.mean(axis=self.axes, keepdims=True)
    array  /= array.std(axis=self.axes, keepdims=True)
    array2 /= array2.std(axis=self.axes, keepdims=True)
    
    array  = self.scipyx.fft.rfftn(array,  axes=self.axes, workers=self.parallel) # release previous array
    array2 = self.scipyx.fft.rfftn(array2, axes=self.axes, workers=self.parallel) # release previous array2
    self.xp.conj(array2, out=array2)
    array *= array2
    del(array2) # release previous array 2
    array = self.scipyx.fft.irfftn(array, axes=self.axes, workers=self.parallel) # release previous array    
    return self.scipyx.fft.fftshift(array, axes=self.axes)



# Main function
vol_crosscorr = FunctionArray(
    
    # Mandatory
    ndims = 3,
    cpu_function = _vol_crosscorr,
    par_function = par_vol_crosscorr,
    gpu_function = _vol_crosscorr,
    out_function = None,
    ini_function = None,

    # Loops
    use_joblib = False, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    func = vol_crosscorr

    # Arguments
    kwargs = dict(
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, narrays=2, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-22
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : transform_matrix

"""
This file allows to test transform_matrix

transform_matrix : Defines the 3x3 transformation matrix for affine transformation in 2D.
"""



# %% Libraries
from corelp import debug
import pytest
from arrlp import transform_matrix
debug_folder = debug(__file__)



# %% Function test
def test_function() :
    '''
    Test transform_matrix function
    '''
    print('Hello world!')



# %% Instance fixture
@pytest.fixture()
def instance() :
    '''
    Create a new instance at each test function
    '''
    return transform_matrix()

def test_instance(instance) :
    '''
    Test on fixture
    '''
    pass


# %% Returns test
@pytest.mark.parametrize("args, kwargs, expected, message", [
    #([], {}, None, ""),
    ([], {}, None, ""),
])
def test_returns(args, kwargs, expected, message) :
    '''
    Test transform_matrix return values
    '''
    assert transform_matrix(*args, **kwargs) == expected, message



# %% Error test
@pytest.mark.parametrize("args, kwargs, error, error_message", [
    #([], {}, None, ""),
    ([], {}, None, ""),
])
def test_errors(args, kwargs, error, error_message) :
    '''
    Test transform_matrix error values
    '''
    with pytest.raises(error, match=error_message) :
        transform_matrix(*args, **kwargs)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-17
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : get_xp

"""
This function returns numpy/cupy depending on the input.
"""



# %% Libraries
import numpy as np
try :
    import cupy as cp
except ImportError :
    cp = None



# %% Function
def get_xp(input) :
    '''
    This function returns numpy/cupy depending on the input.
    
    Parameters
    ----------
    input : bool or array
        If bool, says if using cuda, if array, will detect on cpu or gpu.

    Returns
    -------
    xp : module
        numpy or cupy.

    Examples
    --------
    >>> from arrlp import xp
    ...
    >>> get_xp(use_cuda) # with bool
    >>> get_xp(array) # with array
    '''

    if cp is None :
        return np
    if isinstance(input, bool) :
        return cp if input else np
    return cp.get_array_module(input)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def ini_arr_function(self, array, **kwargs) :
    return dict()

def out_arr_function(self, array, **kwargs) :
    return self.xp.empty_like(array)

def cpu_arr_function(self, out, array, **kwargs) :
    return out # TODO

def par_arr_function(self, out, array, **kwargs) :
    return out # TODO

def gpu_arr_function(self, out, array, **kwargs) :
    return out # TODO



# Main function
arr_function = FunctionArray(
    
    # Mandatory
    ndims = 2,
    cpu_function = cpu_arr_function,
    par_function = par_arr_function,
    gpu_function = gpu_arr_function,
    out_function = out_arr_function,
    ini_function = ini_arr_function,

    # Loops
    cpu_loop = False,
    par_loop = False,
    gpu_loop = False,
    use_joblib = True, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    func = arr_function

    # Arguments
    kwargs = dict(
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-17
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : relabel

"""
This function redefines a label array to fill potential holes inside the considered labels.
"""



# %% Libraries
from arrlp import get_xp



# %% Function
def relabel(labels) :
    '''
    This function redefines a label array to fill potential holes inside the considered labels.
    
    Parameters
    ----------
    labels : np.array
        Array of labels.

    Returns
    -------
    newlabels : np.array
        New array of labels.

    Examples
    --------
    >>> from arrlp import relabel
    ...
    >>> relabel(array)
    '''

    xp = get_xp(labels)
    shape = labels.shape
    labels = labels.ravel().astype(xp.uint32)  # Flatten array
    unique_values = xp.unique(labels)  # Get sorted unique values
    mapping = xp.zeros(unique_values[-1].item() + 1, dtype=xp.uint32)  # Create mapping array
    indices = xp.arange(len(unique_values), dtype=xp.uint32)
    mapping[unique_values] = indices
    labels[:] = mapping[labels]
    return labels.reshape(shape)



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : coordinates

"""
This file allows to test coordinates

coordinates : This function will return a ndarray corresponding to the coordinates array of the input.
"""



# %% Libraries
from corelp import debug
import pytest
import numpy as np
from arrlp import coordinates
debug_folder = debug(__file__)

# %% Returns test
@pytest.mark.parametrize("args, kwargs, expected, message", [
    #([], {}, None, ""),
    ([np.ones((5,4))], {"pixel":10, "center":(True, False), "grid":False, "origin":(0,3)}, (np.array([[-20.], [-10.], [0.], [10.], [20.]]), np.array([[3., 13., 23., 33.]])), ""),
])
def test_returns(args, kwargs, expected, message) :
    '''
    Test coordinates return values
    '''
    assert (coordinates(*args, **kwargs)[0] == expected[0]).all(), message
    assert (coordinates(*args, **kwargs)[1] == expected[1]).all(), message



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def _vol_ifft(self, out, array, **kwargs) :
    return self.scipyx.fft.ifftn(self.scipyx.fft.ifftshift(array, axes=self.axes), axes=self.axes, **kwargs)

def par_vol_ifft(self, out, array, **kwargs) :
    return self.scipyx.fft.ifftn(self.scipyx.fft.ifftshift(array, axes=self.axes), axes=self.axes, workers=self.parallel, **kwargs)



# Main function
vol_ifft = FunctionArray(
    
    # Mandatory
    ndims = 3,
    cpu_function = _vol_ifft,
    par_function = par_vol_ifft,
    gpu_function = _vol_ifft,
    out_function = None,
    ini_function = None,

    # Loops
    use_joblib = False, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    func = vol_ifft

    # Arguments
    kwargs = dict(
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-22
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : transform_matrix

"""
Defines the 3x3 transformation matrix for affine transformation in 2D.
"""



# %% Libraries
import numpy as np



# %% Function
def transform_matrix(shape=None, *, shiftx=0., shifty=0., shearx=0., sheary=0., angle=0., scalex=1., scaley=1.) :
    '''
    Defines the 3x3 transformation matrix for affine transformation in 2D.
    
    Parameters
    ----------
    shape : int or tuple or np.ndarray
        Describes the shape of the coordinates.
        If int, all dimensions will have this value.
        If tuple, corresponds to the shape.
    shiftx : float
        Shift value in x
    shifty : float
        Shift value in y
    shearx : float
        Shear value in x
    sheary : float
        Shear value in y
    angle : float
        Rotation angle [°]
    scalex : float
        Scale value in x
    scaley : float
        Scale value in y

    Examples
    --------
    >>> from arrlp import transform_matrix
    ...
    >>> matrix = transform_matrix(shape, shiftx=2, shifty=3.4)
    '''

    # Manage shape argument
    if shape is None :
        shape = (0, 0)
    elif isinstance(shape, int) :
        shape = (shape,) * 2
    elif isinstance(shape, tuple) :
        pass
    else :
        shape = shape.shape

    h, w = shape
    h, w = h/2, w/2
    matrix = np.eye(3, dtype=np.float32)
    if shearx != 0 or sheary != 0 :
        matrix @= np.asarray([[1., 0., h],
                              [0., 1., w],
                              [0., 0., 1.]])
        matrix @= np.asarray([[1., sheary, 0.],
                              [shearx, 1., 0.],
                              [0., 0., 1.]])
        matrix @= np.asarray([[1., 0., -h],
                              [0., 1., -w],
                              [0., 0., 1.]])
    if angle != 0 :
        angle = -np.radians(angle)
        matrix @= np.asarray([[1., 0., h],
                              [0., 1., w],
                              [0., 0., 1.]])
        matrix @= np.asarray([[np.cos(angle), -np.sin(angle), 0.],
                              [np.sin(angle), np.cos(angle), 0.],
                              [0., 0., 1.]])
        matrix @= np.asarray([[1., 0., -h],
                              [0., 1., -w],
                              [0., 0., 1.]])
    if scalex != 1 or scaley != 1 :
        matrix @= np.asarray([[1., 0., h],
                              [0., 1., w],
                              [0., 0., 1.]])
        matrix @= np.asarray([[scaley, 0., 0.],
                              [0., scalex, 0.],
                              [0., 0., 1.]])
        matrix @= np.asarray([[1., 0., -h],
                              [0., 1., -w],
                              [0., 0., 1.]])
    if shiftx != 0 or shifty != 0 :
        matrix @= np.asarray([[1., 0., shifty],
                              [0., 1., shiftx],
                              [0., 0., 1.]])
    return matrix



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import vol_correlate1d
import numpy as np
try :
    import cupy as cp
except ImportError :
    cp = None



# %% Function
def vol_convolve1d(array, *, kernel, out=None, # Arrays
        stacks=False, channels=False, parallel=False, cuda=False, test=False, iterator=range, # Modes
        **kwargs) :

    xp = cp if cuda and cp is not None else np
    kernel = [xp.flip(k, axis=0) for k in kernel]        
    return vol_correlate1d(array, kernel=kernel, out=out, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, test=test, iterator=iterator, **kwargs)
vol_convolve1d.ndims = 3



if __name__ == '__main__' :
    from arrlp import debug_array
    from arrlp import kernel
    func = vol_convolve1d

    # Arguments
    kwargs = dict(
        kernel=(kernel(ndims=1, pixel=1, sigma=2), kernel(ndims=1, pixel=1, sigma=3), kernel(ndims=1, pixel=1, sigma=3)),
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def ini_vol_minifilter(self, array, *, footprint, **kwargs) :
    return dict(
        footprint=self.xp.asarray(footprint),
    )

def out_vol_minifilter(self, array, **kwargs) :
    return self.xp.empty_like(array, dtype=self.xp.float32)

def _vol_minifilter(self, out, array, *, footprint, mode='nearest', **kwargs) :
    return self.ndimage.minimum_filter(array, footprint=footprint, axes=self.axes, output=out, mode=mode, **kwargs)

def par_vol_minifilter(self, array, *, footprint, mode='nearest', **kwargs) :
    return self.ndimage.minimum_filter(array, footprint=footprint, mode=mode, **kwargs)



# Main function
vol_minifilter = FunctionArray(
    
    # Mandatory
    ndims = 3,
    cpu_function = _vol_minifilter,
    par_function = par_vol_minifilter,
    gpu_function = _vol_minifilter,
    out_function = out_vol_minifilter,
    ini_function = ini_vol_minifilter,

    # Loops
    par_loop = True,
    use_joblib = True, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    from arrlp import kernel
    func = vol_minifilter

    # Arguments
    kwargs = dict(
        footprint=kernel(ndims=3, pixel=1, window=7),
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : kernel

"""
This funtion creates kernels small arrays, mainly for convolutions.
"""



# %% Libraries
from arrlp import coordinates
import numpy as np
try :
    import cupy as cp
except ImportError :
    cp = None
from scipy.special import erf, erfinv, j1



# %% Function
def kernel(ndims=2, pixel=1, *, dtype=np.float32, shape=None, cuda=False, atrou=0, window=None, sigma=None, wl=None, NA=None) :
    '''
    This funtion creates kernels small arrays, mainly for convolutions.
    
    Parameters
    ----------
    ndims : int
        Number of dimensions.
    pixel : int
        Size of pixel for each dimension.
    dtype : np.dtype
        data type of kernel.
    shape : tuple
        shape of kernel.
    cuda : bool
        True to apply cuda.
    atrou : int
        Number of trou to put for atrou algorithm, None for normal kernel.
    window : float
        Size of window for mask kernel (diameter).
    sigma : float
        Size of sigma for gaussian kernel.
    wl : float
        wavelength for Airy kernel.
    NA : float
        Numerical Aperture for Airy kernel.

    Returns
    -------
    k : np.array
        kernel generated.

    Raises
    ------
    SyntaxError
        One keyword input must be put that defines type of kernel, if not will raise Error.

    Examples
    --------
    >>> from arrlp import kernel
    ...
    >>> kernel(window=10) # Mask
    >>> kernel(sigma=5) # Gaussian
    >>> kernel(wl=640, NA=1.5) # Airy
    '''

    #kernel
    if sigma is not None :
        k = kernel_gaus(sigma, pixel, ndims, shape)
    elif window is not None :
        k = kernel_mask(window, pixel, ndims, shape)
    elif wl is not None and NA is not None :
        k = kernel_airy(wl, NA, pixel, ndims, shape)
    else :
        raise SyntaxError('Kernel was not recognized')

    #"à trou" kernel
    if atrou is not None and atrou != 0 :
        k = kernel_atrou(k, atrou)

    k = k.astype(dtype)
    k /= k.sum()
    xp = cp if cuda and cp is not None else np
    return xp.asarray(k)



def kernel_gaus(sigma, pixel=1, ndims=1, shape=None, tol_zero=1/100) :
    '''gets a n dimensions gaussian kernel'''

    # Coordinates
    sigma = np.asarray(iterable(sigma, ndims))
    pixel = np.asarray(iterable(pixel, ndims))
    if shape is None :
        shape = np.ceil(sigma * 6 / pixel)
        shape = tuple((shape + 1 - shape % 2).astype(int)) #odd number shape
    coords = coordinates(shape, center=True, pixel=pixel)

    # Apply Gaussian on each dimension and calculates distance from origin in sigmas
    k = 1
    r2 = 0 # (r/sigma)²
    for coord, sig, pix in zip(coords, sigma, pixel) :
        mini = (coord - pix/2) / np.sqrt(2) / sig
        maxi = (coord + pix/2) / np.sqrt(2) / sig
        k = k * ((erf(mini) - erf(maxi)) / 2)
        r2 = r2 + (coord / sig)**2

    # Apply mask
    mask = r2 > 2 * erfinv((1 - tol_zero)**(1/ndims))**2 # r² > (n**(1/ndims)*sqrt(2))**2
    k[mask] = 0
    return k



def kernel_mask(window, pixel=1, ndims=1, shape=None) :
    '''gets a n dimensions mask kernel'''

    # Coordinates
    window = np.asarray(iterable(window, ndims))
    pixel = np.asarray(iterable(pixel, ndims))
    if shape is None :
        shape = np.ceil(window / pixel)
        shape = tuple((shape + 1 - shape % 2).astype(int)) #odd number shape
    coords = coordinates(shape, center=True, pixel=pixel)

    # Calculates distance from origin in windows
    k = np.ones(shape)
    r2 = 0 # (r/window)²
    for coord, win in zip(coords, window) :
        r2 = r2 + (coord / win)**2

    # Apply mask
    mask = r2 > 1/4 #r² > (1/2)**2
    k[mask] = 0
    return k



def kernel_airy(wl, NA, pixel=1, ndims=1, shape=None) :
    '''gets a n dimensions mask kernel'''

    # Coordinates
    pixel = np.asarray(iterable(pixel, ndims))
    if shape is None :
        shape = np.ceil((2.44 * wl / NA) / pixel)
        shape = tuple((shape + 1 - shape % 2).astype(int)) #odd number shape
    coords = coordinates(shape, center=True, pixel=pixel)

    # Calculates distance from origin in nm and pixel apply Airy function
    r2 = 0 # r²
    for coord in coords :
        r2 = r2 + coord**2
    r = np.sqrt(r2)
    z = 2 * np.pi * r * NA / wl
    ma = r < 1 # < 1 nm
    k = np.empty(shape, dtype=np.float32)
    k[~ma] = (2 * j1(z[~ma]) / z[~ma])**2
    k[ma] = 1

    # Apply mask
    mask = r2 > (1.22 * wl / NA)**2 #r² > (1.22 * wl/NA)**2
    k[mask] = 0
    return k



def kernel_atrou(k, order) :
    if order == 0 :
        return k
    shape = k.shape
    newshape = tuple([s + (s - 1) * order for s in shape])
    newkernel = np.zeros(newshape)
    slices = [slice(None,None,order+1) for _ in range(k.ndim)]
    newkernel[*slices] = k
    return newkernel



def iterable(arg, ndims) :
    try :
        if len(arg) != ndims :
            raise ValueError('Argument does not have good number of dimensions')
        else :
            return arg
    except TypeError :
        return (arg,) * ndims



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-03-18
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : sortloop

"""
This generator allows to loop on unique values of an array rapidly by first sorting the array.
"""



# %% Libraries
from arrlp import get_xp



# %% Function
def sortloop(array, *arrays, axis=None, cuda=False, **kwargs) :
    '''
    This generator allows to loop on unique values of an array rapidly by first sorting the array.
    
    Parameters
    ----------
    array : ndarray
        The array to determine unique values and slices.
    *arrays : ndarrays
        Other arrays to slice in parallel with `array`.
    cuda : bool
        True to use on GPU in cupy.
    axis : int or None
        Axis along which to find unique values. If None, flattens the array.

    Returns
    -------
    positions : array
        Positions corresponding to the current masked elements.
    array : array
        first array which was used for defining the masked region
    *arrays :
        other array elements in the masked region

    Examples
    --------
    >>> from arrlp import sortloop
    ...
    >>> for pos, r, y, x in sortloop(R, Y, X) :
    ...     result[pos] = scalar_function(y, x) # masked for each r
    '''

    # arrays
    xp = get_xp(cuda)
    array = xp.asarray(array)
    arrays = [xp.asarray(a) for a in arrays]
    
    # axis
    if axis is None:
        flat_array = array.ravel()
        flat_arrays = [a.ravel() for a in arrays]
        ax = 0
    else:
        flat_array = array
        flat_arrays = list(arrays)
        ax = axis

    # sort
    argsort = xp.argsort(flat_array, axis=ax)
    sorted_array = flat_array[argsort]
    sorted_arrays = [a[argsort] for a in flat_arrays]

    # unique values and counts
    unique, counts = xp.unique(sorted_array, return_counts=True, axis=ax)
    cumsum = xp.hstack((0, xp.cumsum(counts)))

    # generate chunks
    for i, val in enumerate(unique):
        slc = slice(cumsum[i], cumsum[i+1])
        positions = argsort[slc]
        chunk_arrays = [a[slc] for a in sorted_arrays]
        yield positions, val, *chunk_arrays



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray, coordinates



# %% Function

# Initializations
def ini_img_radialproj(self, array, *, pixel=1., **kwargs) :
    try :
        if len(pixel) != 2 :
            raise ValueError('pixel cannot have other than 2 values in tuple')
    except TypeError :
        pixel = pixel, pixel
    ratio = pixel[0] / min(pixel), pixel[1]/min(pixel)
    y, x = coordinates(self.shape(array), pixel=ratio, grid=True, cuda=self.cuda)
    R = (self.xp.sqrt(y**2 + x**2)).astype(int).ravel()
    rsum = self.xp.bincount(R.ravel())
    return dict(R=R, rsum=rsum) # return values in this dict will be passed as kwargs in cpu/par/gpu functions 

def _img_radialproj(self, out, array, *, R, rsum, **kwargs) :
    psum = self.xp.bincount(R, weights=array.ravel())
    return psum / rsum

def par_img_radialproj(self, array, *, R, rsum, **kwargs) :
    psum = self.xp.bincount(R, weights=array.ravel())
    return psum / rsum



# Main function
img_radialproj = FunctionArray(
    
    # Mandatory
    ndims = 2,
    cpu_function = _img_radialproj,
    par_function = par_img_radialproj,
    gpu_function = _img_radialproj,
    out_function = None,
    ini_function = ini_img_radialproj,

    # Loops
    cpu_loop = True,
    par_loop = True,
    gpu_loop = True, # True usually make bad GPU performance, check first if there is no ndimension support, and also try investigating reshaping
    use_joblib = True, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = True,
    remove_cuda = True,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    func = img_radialproj

    # Arguments
    kwargs = dict(
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-09
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : signal

"""
Signal array functions of base shape (x,).
"""



# %% Libraries
from arrlp import *



# %% Function
def signal(**kwargs) :
    '''
    Signal array functions of base shape (x,).
    
    Parameters
    ----------
    a : int or float
        TODO.

    Returns
    -------
    b : int or float
        TODO.

    Raises
    ------
    TypeError
        TODO.

    Examples
    --------
    >>> from arrlp import signal
    ...
    >>> signal() # TODO
    '''

    return None



# %% Libraries
from corelp import prop
from arrlp import *
from dataclasses import dataclass, field



# %% Class
@dataclass(slots=True, kw_only=True)
class signal() :
    '''
    Signal array functions of base shape (x,).
    
    Parameters
    ----------
    a : int or float
        TODO.

    Attributes
    ----------
    _attr : int or float
        TODO.

    Examples
    --------
    >>> from arrlp import signal
    ...
    >>> instance = signal(TODO)
    '''

    # Attributes
    # myattr : str = ""
    # mylist : list[str] = field(default_factory=list)
    # _mycal : str = field(init=False, repr=False) 
    name : str = None



    # Init
    def __post_init__(self) :
        pass



    # Properties
    @prop(cache=True)
    def myprop(self) :
        return ""
    @myprop.setter()
    def mypropr(self, value) :
        self._myprop = value



    # Methods
    def method(self) :
        '''
        TODO
    
        Parameters
        ----------
        a : int or float
            TODO.

        Returns
        -------
        b : int or float
            TODO.

        Raises
        ------
        TypeError
            TODO.

        Examples
        --------
        >>> self.method() # TODO
        '''

        return None



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def ini_img_transform(self, array, *, matrix, **kwargs) :
    return dict(
        matrix=self.xp.asarray(matrix),
    )

def out_img_transform(self, array, **kwargs) :
    return self.xp.empty_like(array, dtype=self.xp.float32)

def _img_transform(self, out, array, *, matrix, **kwargs) :
    return self.ndimage.affine_transform(array, matrix[:2, :2], offset=matrix[:2, 2], output=out, **kwargs)

def par_img_transform(self, array, *, matrix, **kwargs) :
    return self.ndimage.affine_transform(array, matrix[:2, :2], offset=matrix[:2, 2], **kwargs)



# Main function
img_transform = FunctionArray(
    
    # Mandatory
    ndims = 2,
    cpu_function = _img_transform,
    par_function = par_img_transform,
    gpu_function = _img_transform,
    out_function = out_img_transform,
    ini_function = ini_img_transform,

    # Loops
    cpu_loop = True,
    par_loop = True,
    gpu_loop = True, # True usually make bad GPU performance, check first if there is no ndimension support, and also try investigating reshaping
    use_joblib = True, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = False,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    from arrlp import transform_matrix
    func = img_transform

    # Arguments
    kwargs = dict(
        matrix=transform_matrix(shiftx=2.3),
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-28
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP

"""
A library that provides tools for arrays on CPU & GPU [any shape].
"""



# %% Source import
sources = {
'FunctionArray': 'arrlp.modules.FunctionArray_LP.FunctionArray',
'compress': 'arrlp.modules.compress_LP.compress',
'coordinates': 'arrlp.modules.coordinates_LP.coordinates',
'debug_array': 'arrlp.modules.debug_array_LP.debug_array',
'gc': 'arrlp.modules.gc_LP.gc',
'get_xp': 'arrlp.modules.get_xp_LP.get_xp',
'image': 'arrlp.modules.image_LP.image',
'kernel': 'arrlp.modules.kernel_LP.kernel',
'nb_threads': 'arrlp.modules.nb_threads_LP.nb_threads',
'relabel': 'arrlp.modules.relabel_LP.relabel',
'signal': 'arrlp.modules.signal_LP.signal',
'sortloop': 'arrlp.modules.sortloop_LP.sortloop',
'transform_matrix': 'arrlp.modules.transform_matrix_LP.transform_matrix',
'volume': 'arrlp.modules.volume_LP.volume',
'img_ifft': 'arrlp.modules.image_LP._functions.img_ifft',
'img_autocorr': 'arrlp.modules.image_LP._functions.img_autocorr',
'img_gaussianfilter': 'arrlp.modules.image_LP._functions.img_gaussianfilter',
'img_convolve': 'arrlp.modules.image_LP._functions.img_convolve',
'img_crosscorr': 'arrlp.modules.image_LP._functions.img_crosscorr',
'img_': 'arrlp.modules.image_LP._functions.img_',
'img_wiener': 'arrlp.modules.image_LP._functions.img_wiener',
'img_maxifilter': 'arrlp.modules.image_LP._functions.img_maxifilter',
'img_fft': 'arrlp.modules.image_LP._functions.img_fft',
'img_radialproj': 'arrlp.modules.image_LP._functions.img_radialproj',
'img_minifilter': 'arrlp.modules.image_LP._functions.img_minifilter',
'img_convolve1d': 'arrlp.modules.image_LP._functions.img_convolve1d',
'img_correlate': 'arrlp.modules.image_LP._functions.img_correlate',
'img_greyopening': 'arrlp.modules.image_LP._functions.img_greyopening',
'img_correlate1d': 'arrlp.modules.image_LP._functions.img_correlate1d',
'img_transform': 'arrlp.modules.image_LP._functions.img_transform',
'sig_maxifilter': 'arrlp.modules.signal_LP._functions.sig_maxifilter',
'sig_wiener': 'arrlp.modules.signal_LP._functions.sig_wiener',
'sig_correlate': 'arrlp.modules.signal_LP._functions.sig_correlate',
'sig_ifft': 'arrlp.modules.signal_LP._functions.sig_ifft',
'sig_gaussianfilter': 'arrlp.modules.signal_LP._functions.sig_gaussianfilter',
'sig_greyopening': 'arrlp.modules.signal_LP._functions.sig_greyopening',
'sig_minifilter': 'arrlp.modules.signal_LP._functions.sig_minifilter',
'sig_convolve': 'arrlp.modules.signal_LP._functions.sig_convolve',
'sig_crosscorr': 'arrlp.modules.signal_LP._functions.sig_crosscorr',
'sig_autocorr': 'arrlp.modules.signal_LP._functions.sig_autocorr',
'sig_fft': 'arrlp.modules.signal_LP._functions.sig_fft',
'vol_crosscorr': 'arrlp.modules.volume_LP._functions.vol_crosscorr',
'vol_ifft': 'arrlp.modules.volume_LP._functions.vol_ifft',
'vol_gaussianfilter': 'arrlp.modules.volume_LP._functions.vol_gaussianfilter',
'vol_correlate': 'arrlp.modules.volume_LP._functions.vol_correlate',
'vol_maxifilter': 'arrlp.modules.volume_LP._functions.vol_maxifilter',
'vol_fft': 'arrlp.modules.volume_LP._functions.vol_fft',
'vol_convolve1d': 'arrlp.modules.volume_LP._functions.vol_convolve1d',
'vol_correlate1d': 'arrlp.modules.volume_LP._functions.vol_correlate1d',
'vol_wiener': 'arrlp.modules.volume_LP._functions.vol_wiener',
'vol_autocorr': 'arrlp.modules.volume_LP._functions.vol_autocorr',
'vol_minifilter': 'arrlp.modules.volume_LP._functions.vol_minifilter',
'vol_greyopening': 'arrlp.modules.volume_LP._functions.vol_greyopening',
'vol_convolve': 'arrlp.modules.volume_LP._functions.vol_convolve'
}



# %% Lazy imports
from corelp import getmodule
__getattr__, __all__ = getmodule(sources)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-09
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : image

"""
Image array functions of base shape (y, x).
"""



# %% Libraries




# %% Function
def image(**kwargs) :
    '''
    Image array functions of base shape (y, x).
    
    Parameters
    ----------
    a : int or float
        TODO.

    Returns
    -------
    b : int or float
        TODO.

    Raises
    ------
    TypeError
        TODO.

    Examples
    --------
    >>> from arrlp import image
    ...
    >>> image() # TODO
    '''

    return None



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : coordinates

"""
This function will return a ndarray corresponding to the coordinates array of the input.
"""



# %% Libraries
import numpy as np
from arrlp import get_xp



# %% Function
def coordinates(shape, pixel=1., *, ndims=1, center=True, grid=False, origin=0., cuda=False) :
    '''
    This function will return a ndarray corresponding to the coordinates array of the input.
    
    Parameters
    ----------
    shape : int or tuple or np.ndarray
        Describes the shape of the coordinates.
        If int, all dimensions will have this value.
        If tuple, corresponds to the shape.
    pixel : float, tuple(float)
        Says what is the size of one bin, default is 1.
        If tuple, corresponds to the pixel for each dimension.
    ndims : int
        Number of dimensions if input shape is int.
        Will be overriden by any other input dimension.
    center : bool or tuple(bool)
        True to center origin, False to start by origin.
        If tuple, defines how to place origin for each dimension.
    grid : bool
        True to return full meshgrid for each dimensions, False for broadcastable arrays.
    origin : float or tuple(float)
        Value of origin (default 0).
        If tuple, value for each dimension.
    cuda : bool
        True to apply cuda.

    Returns
    -------
    coords : tuple(np.ndarray)
        return a tuple of arrays for each dimensions.

    Raises
    ------
    ValueError
        if input or passed as tuples that do not have the good number of dimensions.

    Examples
    --------
    >>> from arrlp import coordinates
    ...
    >>> array = np.ones((5,4))
    >>> coordinates = coordinates(array, pixel=10, center=(True, False), grid=False, origin=(0,3))
    ... array([[-20.], [-10.], [0.], [10.], [20.]]), array([[3., 13., 23., 33.]])
    '''
    xp = get_xp(cuda)

    # Manage shape argument
    if isinstance(shape, int) :
        shape = (shape,) * ndims
    elif isinstance(shape, tuple) :
        pass
    else :
        shape = shape.shape

    # Correct ndims
    ndims = len(shape)

    # Correct input dimensions
    pixel = iterable(pixel, ndims)
    center = iterable(center, ndims)
    origin = iterable(origin, ndims)

    #looping on shape
    coords = []
    for dim, (n, pix, cent, orig) in enumerate(zip(shape, pixel, center, origin)) :
        coord = n2coord(n, pixel=pix, center=cent, origin=orig, xp=xp)
        reshape = np.ones(ndims, dtype=int)
        reshape[dim] = len(coord)
        coord = coord.reshape(tuple(reshape))
        coords.append(coord)

    #Meshgrid
    if grid and ndims != 1 :
        return xp.meshgrid(*[coord.squeeze() for coord in coords], indexing='ij')
    else :
        return coords



def iterable(arg, ndims) :
    try :
        if len(arg) != ndims :
            raise ValueError('Argument does not have good number of dimensions')
        else :
            return arg
    except TypeError :
        return (arg,) * ndims



def n2coord(n, pixel=1., center=False, origin=0., xp=np) :
    if center :
        start, stop = -(n - 1) / 2,(n - 1) / 2
    else :
        start, stop = 0, n - 1
    return xp.linspace(start, stop, n) * pixel + origin



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : kernel

"""
This file allows to test kernel

kernel : This funtion creates kernels small arrays, mainly for convolutions.
"""



# %% Libraries
from corelp import debug
import pytest
from matplotlib import pyplot as plt
import numpy as np
from arrlp import kernel
debug_folder = debug(__file__)



# %% Function test
@pytest.mark.parametrize("kwargs, title", [
    ({"window":500}, "Mask"),
    ({"sigma":250}, "Gaussian"),
    ({"wl":640, "NA":1.5}, "Airy"),
])
def test_function(kwargs, title) :
    '''
    Test kernel values
    '''
    k = kernel(pixel=10, **kwargs)
    assert np.isclose(k.sum(), np.float32(1.))
    plt.figure()
    plt.imshow(k, cmap="hot")
    plt.tight_layout()
    plt.colorbar()
    plt.savefig(debug_folder / f"{title}.png")



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import FunctionArray



# %% Function

# Initializations
def ini_sig_wiener(self, array, *, kernel, power=2, balance=10**2, **kwargs) :
    kernel = self.xp.asarray(kernel)
    axes = self.axes
    X = array.shape[axes[0]]
    pad_x_total = X - kernel.shape[0]
    if pad_x_total < 0 :
        start = -pad_x_total // 2
        kernel = kernel[start:start + X]
        pad_x_total = 0
    pad_x = (pad_x_total // 2, pad_x_total - pad_x_total // 2)
    padded = self.xp.pad(kernel, pad_x, mode='constant')
    fft_kernel = self.xp.abs(self.scipyx.fft.fftshift(self.scipyx.fft.fft(padded)))
    W = fft_kernel**(power-1) / (fft_kernel**power + fft_kernel.max() / balance)
    if self.stacks and self.channels :
        W = W.reshape((1, *W.shape, 1))
    elif self.stacks :
        W = W.reshape((1, *W.shape))
    elif self.channels :
        W = W.reshape((*W.shape, 1))
    return dict(
        W=W,
    )

def _sig_wiener(self, out, array, *, W, kernel, power=2, balance=10**2, **kwargs) :
    dtype = array.dtype
    fft = self.scipyx.fft.fftshift(self.scipyx.fft.fft(array, axis=self.axes[0], **kwargs), axes=self.axes)
    return self.xp.real(self.scipyx.fft.ifft(self.scipyx.fft.ifftshift(fft * W, axes=self.axes), axis=self.axes[0], **kwargs)).astype(dtype)

def par_sig_wiener(self, out, array, *, W, kernel, power=2, balance=10**2, **kwargs) :
    dtype = array.dtype
    fft = self.scipyx.fft.fftshift(self.scipyx.fft.fft(array, axis=self.axes[0], workers=self.parallel, **kwargs), axes=self.axes)
    return self.xp.real(self.scipyx.fft.ifft(self.scipyx.fft.ifftshift(fft * W, axes=self.axes), axis=self.axes[0], workers=self.parallel, **kwargs)).astype(dtype)



# Main function
sig_wiener = FunctionArray(
    
    # Mandatory
    ndims = 1,
    cpu_function = _sig_wiener,
    par_function = par_sig_wiener,
    gpu_function = _sig_wiener,
    out_function = None,
    ini_function = ini_sig_wiener,

    # Loops
    use_joblib = False, # If True, arguments of parallel function should not have "out".

    # Performances
    remove_parallel = True,
    remove_cuda = False,

)



if __name__ == '__main__' :
    from arrlp import debug_array
    from arrlp import kernel
    func = sig_wiener

    # Arguments
    kwargs = dict(
        kernel = kernel(ndims=1, sigma=3),
    )

    # Modes
    modes = { # list of dicts with kwargs
        "Reference": dict(cuda=False, parallel=False),
        "Parallel": dict(cuda=False, parallel=True),
        "Cuda": dict(cuda=True, parallel=False),
    }

    debug_array(func, modes, **kwargs)
